/*
 * $Id: Server.java 364 2007-04-09 03:52:26Z davidb $
 *
 * Copyright (c) 2004 USC Project SRO. All rights reserved.
 * Copyright (c) 2005 VeriSign, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

package se.rfc.unbound.server;

import java.io.IOException;
import java.net.*;
import java.nio.channels.*;
import java.util.*;

import org.apache.log4j.Logger;
import org.xbill.DNS.*;
import org.xbill.DNS.utils.hexdump;

import se.rfc.unbound.*;
import se.rfc.unbound.Resolver;
import se.rfc.unbound.iter.IterativeResolver;
import se.rfc.unbound.nsclient.NameserverClient;
import se.rfc.unbound.validator.ValidatingResolver;

/**
 * 
 * 
 * @author davidb
 * @version $Revision: 364 $
 */
public class Server implements Resolver
{

  // Configuration parameters.
  /** This is a particular IP and/or port to listen on, UDP and TCP. */
  private InetSocketAddress    mListenAddress;
  /**
   * This is a particular IP and/or port to send outbound requests via. It
   * will be null if random source ports are desired.
   */
  private InetSocketAddress    mQuerySource;
  /**
   * The precreated UDP channel for the fixed query source, if there is one.
   * This will be null if not.
   */
  private DatagramChannel      mQuerySourceChannel;

  // Server state
  /** The next resolver in the chain -- requests are forwarded to this module. */
  private Resolver             mNextResolver;
  /**
   * The previous resolver in the chain (forming a circular list). Responses
   * are forwarded to this module.
   */
  private Resolver             mPrevResolver;

  private NameserverClient     mNSClient;

  /**
   * This is the main selector -- all active channels are registered with this
   * selector.
   */
  private Selector             mSelector;
  /**
   * This is a set of channels that are listening for incoming requests. If a
   * request is received on a channel not in this set, it will be discarded.
   */
  private Set                  mServerChannelSet;
  /**
   * This is a set of channels that are listening for incoming responses. If a
   * response is received on a channel not in this set, it will be discarded.
   */
  private Set                  mClientChannelSet;
  /**
   * When this gets set to true, the main service loop will exit after one
   * more round.
   */
  private boolean              mStop;
  /**
   * This is a data structure tracking received, active requests. It is used
   * to prevent duplicate efforts and to detect actual duplicate packets.
   */
  private ConcurrentRequestMap mConcurrentRequestMap;

  /** A local reference to a set of utility functions for this module. */
  private ServerUtils          mServerUtils;

  // Logs.
  private Logger               log           = Logger.getLogger(this
                                                 .getClass());
  private Logger               servTrafLog   = Logger
                                                 .getLogger("server_traffic_log");
  private Logger               clientTrafLog = Logger
                                                 .getLogger("client_traffic_log");

  /**
   * Create a server module instance. Generally, there is only ever one of
   * this instantiated at a time.
   */
  public Server()
  {
    mServerChannelSet = new HashSet();
    mClientChannelSet = new HashSet();
    mConcurrentRequestMap = new ConcurrentRequestMap();
  }

  /**
   * Initialize this module.
   * 
   * @param config The resolver configuration.
   * 
   * @throws Exception if something goes wrong with the initialization with
   *           this module, or any chained modules.
   */
  public void init(Properties config) throws Exception
  {
    mServerUtils = new ServerUtils(config);

    String s = null;

    // Set up listen address
    // FIXME: multiple listen addresses? maybe for more than one interface,
    // but not all?
    InetAddress addr = null;
    s = config.getProperty("server.address");

    if (s != null)
    {
      try
      {
        addr = InetAddress.getByName(s);
      }
      catch (UnknownHostException e)
      {
        log.error("could not resolve server.address: " + s, e);
      }
    }

    // Set up listen port
    s = config.getProperty("server.port");
    int port = Util.parseInt(s, 53);

    mListenAddress = new InetSocketAddress(addr, port);

    // Set up query source port/address.
    // FIXME: should be able to have a query_source_address, too.
    s = config.getProperty("server.query_source_port");
    int qport = Util.parseInt(s, -1);
    if (qport > 0)
    {
      mQuerySource = new InetSocketAddress(qport);
    }

    s = config.getProperty("server.server_selection.class",
        "se.rfc.unbound.SimpleServerSelection");
    ServerSelection serv_sel = ServerUtils.loadServerSelection(s);

    // FIXME: for now, this is hardcoded.

    // The NameserverClient module must be initialized with a reference to
    // this module, to put it in asynchronous mode.
    mNSClient = new NameserverClient(this, serv_sel);
    mNextResolver = new ValidatingResolver(new IterativeResolver(mNSClient,
        serv_sel));
    setPrevResolver(mNSClient);
    mNextResolver.setPrevResolver(this);

    mNextResolver.init(config);
    mNextResolver.setResolutionHead(this);
  }

  /**
   * This is just a convenience routine for getting the module state for this
   * module from an event.
   * 
   * @return The module state object for this module, or null if one hasn't
   *         been attached yet.
   */
  private ServerEventState getModuleState(DNSEvent event)
  {
    return (ServerEventState) event.getModuleState(this);
  }

  /**
   * Shutdown this resolver module. This will be forwarded to the next module
   * in the chain.
   */
  public void shutdown()
  {
    mNextResolver.shutdown();
  }

  /**
   * This is the main request entry point. Since this is a standard chain
   * head, all received and generated queries should pass through this point.
   */
  public void processRequest(DNSEvent event)
  {
    // Not much to do here -- all socket generated requests will have been
    // fully vetted by this point.
    // We just pass the request event on to the next resolver.

    mNextResolver.processRequest(event);
  }

  /**
   * This the the main response entry point. This is basically the last stop
   * for final responses before being scheduled for writing.
   * 
   * Responses that have attached concurrent events will have those events
   * processed as well. The effect should be that once the original response
   * is received, it is sent back to the client, then is copied and sent back
   * to the clients of the concurrent requests.
   */
  public void processResponse(DNSEvent event)
  {
    ServerEventState state = getModuleState(event);

    handleWriteResponse(event, state);

    // process any concurrent events here.
    if (state.hasConcurrentEvents())
    {
      List l = state.getConcurrentEvents();
      for (Iterator i = l.iterator(); i.hasNext();)
      {
        DNSEvent c_event = (DNSEvent) i.next();
        ServerEventState c_state = getModuleState(c_event);
        handleWriteResponse(c_event, c_state);
      }
      state.clearConcurrentEvents();
    }

    // Remove this event from the concurrent request map -- it is no longer an
    // active request.
    mConcurrentRequestMap.remove(event);

    // DEBUG: uncomment the next line if you want the server to only handle
    // one request.
    // mStop = true;
  }

  /**
   * Write (actually, schedule a write event) a final response.
   * 
   * @param event The event containing the final response to a received query.
   * @param state The event state, already extracted from the event.
   */
  private void handleWriteResponse(DNSEvent event, ServerEventState state)
  {
    // Basically, we just need to send the response to originator.
    if (state == null)
    {
      log.error("processing response event with no module state!!!");
      return;
    }

    Response response = event.getResponse();
    if (response == null) return;
    SMessage resp = response.getSMessage();
    if (resp == null) return;

    Message m = mServerUtils.formatResponseMessage(resp, event
        .getOrigRequest());
    servTrafLog.debug("sending final response:\n" + m);
    writeResponse(m,
        state.getMaximumPayload(),
        state.getSourceAddress(),
        state.getSourceChannel());
  }

  /**
   * @return The "next" resolver. Next is the direction that requests move in.
   */
  public Resolver getNextResolver()
  {
    return mNextResolver;
  }

  /**
   * @return the "previous" resolver. Previous is the direction that responses
   *         move in.
   */
  public Resolver getPrevResolver()
  {
    return mPrevResolver;
  }

  /**
   * Set the value of the previous resolver.
   */
  public void setPrevResolver(Resolver prev)
  {
    mPrevResolver = prev;
  }

  /**
   * Set the value of the "head" of the resolution chain. The head is where
   * all requests start (officially).
   */
  public void setResolutionHead(Resolver head)
  {
    mNextResolver.setResolutionHead(head);
  }

  /**
   * Endlessly process incoming packets.
   * 
   * @throws IOException
   */
  public void serve() throws IOException
  {
    // initialize selector with initial server sockets
    initializeSelector();

    boolean stop = false;

    long timeout = 0;

    while (true)
    {
      // Block on select, with a timeout. The timeout is used so we can be
      // guaranteed to do some housecleaning events, like timing out requests.

      if (timeout < 0)
      {
        mSelector.selectNow();
      }
      else
      {
        mSelector.select(timeout);
      }

      log.trace("selection round begin");

      for (Iterator i = mSelector.selectedKeys().iterator(); i.hasNext();)
      {
        SelectionKey key = (SelectionKey) i.next();

        // Not certain if this is useful, but it shouldn't hurt.
        if (!key.isValid()) continue;

        ChannelState state = (ChannelState) key.attachment();

        if (key.isAcceptable())
        {
          handleAccept(key, state);
        }

        if (key.isConnectable())
        {
          handleConnect(key, state);
        }

        // The only combined states will be OP_READ | OP_WRITE

        if (key.isReadable())
        {
          // Note that this can render a key invalid (by closing its channel).
          handleRead(key, state);
        }

        if (key.isValid() && key.isWritable())
        {
          // Note that this can render a key invalid (by closing its channel).
          handleWrite(key, state);
        }

        i.remove();
      }

      log.trace("selection round end");

      // Timeout/Wakeup outstanding requests.

      timeout = mNSClient.handleTimeouts();

      log.trace("timeout = " + timeout);

      // DEBUG: this is a two-pass stop. That is, the first pass just set a
      // flag, the second pass actually stops. This is so any final writes may
      // be done. NOTE: this is just for debugging. If an actual orderly
      // shutdown is desired (likely), then this should be modified to try and
      // get all of the final response writes completed, while ceasing to
      // accept any more requests.
      if (stop) return;
      if (mStop) stop = true;
    }
  }

  /**
   * A convenience routine for creating DatagramChannels and registering them
   * with the selector.
   * 
   * @param sourceAddress The source address to bind to; may be null.
   * @param permanent if true, then this channel will be marked as permanent
   *          in it's selection key, so subsequent read/write activies do not
   *          deregister this channel.
   * @param outbound if true, then this channel will be used to send requests
   *          and receive responses. Non-permanent channels will be
   *          deregistered (closed) after they finish one read (if outbound)
   *          or one write (if not).
   * @return The registered channel.
   * @throws IOException
   */
  private DatagramChannel registerNewUDPChannel(
      InetSocketAddress sourceAddress, boolean permanent, boolean outbound)
      throws IOException
  {
    DatagramChannel ch = DatagramChannel.open();
    ch.configureBlocking(false);
    if (sourceAddress != null)
    {
      DatagramSocket sock = ch.socket();
      sock.bind(sourceAddress);
    }

    ChannelState state = new ChannelState(permanent, outbound);
    int ops = outbound ? SelectionKey.OP_WRITE : SelectionKey.OP_READ;
    ch.register(mSelector, ops, state);

    if (outbound)
    {
      mClientChannelSet.add(ch);
    }
    else
    {
      mServerChannelSet.add(ch);
    }

    return ch;
  }

  /**
   * A convenience routine for registering an already created SocketChannel.
   * The channel will be set to non-blocking and added to the selector (and
   * other collections).
   * 
   * @param channel The channel to register.
   * @param permanent If true, then this socket will not be deregistered.
   * @param outbound If true then this socket is meant to send a request and
   *          receive a response.
   * @throws IOException
   */
  private void registerTCPChannel(SocketChannel channel, boolean permanent,
      boolean outbound) throws IOException
  {
    ChannelState state = new ChannelState(permanent, outbound);

    channel.configureBlocking(false);

    int ops = outbound ? SelectionKey.OP_WRITE : SelectionKey.OP_READ;

    channel.register(mSelector, ops, state);

    if (outbound)
    {
      mClientChannelSet.add(channel);
    }
    else
    {
      mServerChannelSet.add(channel);
    }
    log.trace("new TCP channel to: "
        + channel.socket().getRemoteSocketAddress());
  }

  /**
   * Close and remove a channel. This will remove the channel from the
   * appropriate set.
   * 
   * @param channel The channel to close.
   * @throws IOException
   */
  private void deregisterChannel(SelectableChannel channel)
      throws IOException
  {
    SelectionKey key = channel.keyFor(mSelector);
    ChannelState state = (ChannelState) key.attachment();

    if (state.isPermanent()) return;

    // Note that non-permanent channels are only ever in one set or the other,
    // not both (permanent channels might be in both).
    if (state.isOutbound())
    {
      mClientChannelSet.remove(channel);
    }
    else
    {
      mServerChannelSet.remove(channel);
    }

    channel.close();
  }

  /**
   * Create and add the permanent channels to the global selector.
   * 
   * @throws IOException if something goes wrong creating a channel,
   *           typically.
   */
  private void initializeSelector() throws IOException
  {
    mSelector = Selector.open();

    // Configure server UDP channel
    DatagramChannel dch = registerNewUDPChannel(mListenAddress, true, false);
    log.debug("Listening on UDP " + mListenAddress);

    // Configure server TCP channel
    ServerSocketChannel sch = ServerSocketChannel.open();
    sch.configureBlocking(false);
    ServerSocket ssock = sch.socket();
    ssock.bind(mListenAddress);
    sch.register(mSelector, SelectionKey.OP_ACCEPT, new ChannelState(true,
        false));
    mServerChannelSet.add(sch);
    log.debug("Listening on TCP " + mListenAddress);

    // If we have a query source port, and it is different than the server
    // port, configure a long term outbound UDP channel
    if (mQuerySource != null)
    {
      if (mQuerySource.getPort() == mListenAddress.getPort())
      {
        mClientChannelSet.add(dch);
        mQuerySourceChannel = dch;
        log.debug("using " + mListenAddress + " as query source");
      }
      else
      {
        mQuerySourceChannel = registerNewUDPChannel(mQuerySource, true, false);
        log.debug("using " + mQuerySource + " as query source");
      }
    }
  }

  /**
   * Process a channel that is ready to accept -- this only applies to the
   * ServerSocketChannels. Basically this will create a new SocketChannel to
   * handle the incoming TCP connection.
   * 
   * @param key The ready key.
   * @param state The state for this key, not really used here.
   * @throws IOException
   */
  private void handleAccept(SelectionKey key, ChannelState state)
      throws IOException
  {
    log.trace("handling accept");
    ServerSocketChannel sch = (ServerSocketChannel) key.channel();

    // Accept the new TCP channel
    SocketChannel ch = sch.accept();
    // register the channel with the selector + client or server sets.
    registerTCPChannel(ch, false, false);
    log.trace("accepted new TCP channel: " + ch);
  }

  /**
   * Given a key in the connectable state, attempt to finish connecting.
   * 
   * @param key The key in question.
   * @param state The channel state for that key.
   * @throws IOException
   */
  private void handleConnect(SelectionKey key, ChannelState state)
      throws IOException
  {
    log.trace("handling connect");
    SocketChannel sch = (SocketChannel) key.channel();

    boolean res = sch.finishConnect();
    if (!res) return;

    log.trace("outbound TCP channel connection complete: "
        + sch.socket().getRemoteSocketAddress());

    // Otherwise, reset the interest ops
    key.interestOps(SelectionKey.OP_WRITE);
  }

  /**
   * Read, parse, and hand-off DNS messages arriving on a given channel. After
   * reading a DNS message, it attempt to parse the message. If that fails, it
   * will send a FORMERR response, if it can be determined that the malformed
   * message was a request.
   * 
   * If the message does parse, it classified as a request or a response and
   * handed off to the appropriate routine (handleReadRequest() vs.
   * handleReadResponse()).
   * 
   * @param key The key in the readable state.
   * @param state The channel state attached to the key.
   * @throws IOException
   */
  private void handleRead(SelectionKey key, ChannelState state)
      throws IOException
  {
    // First, attempt to read a DNS message from the channel.

    SelectableChannel channel = key.channel();
    ServerUtils.ReadResult rr = null;
    if (channel instanceof DatagramChannel)
    {
      DatagramChannel dch = (DatagramChannel) channel;
      rr = mServerUtils.readUDP(dch);
      if (rr == null)
      {
        return;
      }
    }
    else
    {
      SocketChannel sch = (SocketChannel) channel;
      rr = mServerUtils.readTCP(state, sch);
      if (rr == null)
      {
        // incomplete read
        return;
      }
      // TCP channel closed on us
      if (rr.closeChannel)
      {
        deregisterChannel(channel);
      }
      if (rr.in == null)
      {
        // no complete message was read, so drop it.
        return;
      }
    }

    // Attempt to parse the input buffer as a DNS message
    Message m = parseMessage(rr.in);

    // If the parse attempt failed, and this is a server channel.
    if (m == null)
    {
      // Attempt to parse the header, to try and determine if this garbage was
      // a request.
      try
      {
        Header header = null;
        if (mServerChannelSet.contains(channel)) header = new Header(rr.in);
        if (header != null && !header.getFlag(Flags.QR))
        {
          // input buffer failed to parse.
          writeResponse(mServerUtils.errorMessage(header, Rcode.FORMERR),
              512,
              rr.sourceAddress,
              channel);
          log.debug("incoming request failed to parse from: "
              + rr.sourceAddress);
        }
      }
      catch (WireParseException e)
      {
        log.debug("incoming request failed to parse (at all) from: " + rr.sourceAddress);
      }
        
      // If the header wasn't parseable, or was a response, we just drop it.
      return;
    }

    boolean isResponse = m.getHeader().getFlag(Flags.QR);

    if (isResponse)
    {
      handleReadResponse(m, channel, rr.sourceAddress, state);
    }
    else
    {
      handleReadRequest(m, channel, rr.sourceAddress, state);
    }
  }

  /**
   * Given a request, make sure it came on a valid channel, validate the
   * request (i.e., check to see if it has a valid format and a sensical
   * query), check to see if it is a duplicate (exact or not), and, if not,
   * generate a request event and process it.
   * 
   * @param m The read message.
   * @param channel The channel the message was read from (needed to track
   *          where to send the response from)
   * @param sourceAddress The source address of the message (needed to track
   *          where to send the response to)
   * @param state The channel state attached to the channel.
   */
  private void handleReadRequest(Message m, SelectableChannel channel,
      InetSocketAddress sourceAddress, ChannelState state)
  {
    // First determine if the request came on a proper channel.
    if (!mServerChannelSet.contains(channel))
    {
      // Drop requests arriving on response-only channels.
      log.info("request came on response-only channel, dropping.");
      return;
    }
    log.trace("received request from " + sourceAddress + ":\n" + m);
    servTrafLog.debug("received request from " + sourceAddress + ":\n" + m);

    // Vet the request.
    Message err = mServerUtils.validateRequestMessage(m);
    if (err != null)
    {
      // If the request failed (or is some other server-response capable
      // request), write the response and we are done.
      log.debug("request was not valid");
      writeResponse(err, 512, sourceAddress, channel);
      return;
    }

    // Otherwise, create a new request event.
    DNSEvent request_event = createRequestEvent(m, channel, sourceAddress);

    // Check it for duplication/register it as an active request.
    DNSEvent duplicate = mConcurrentRequestMap
        .getMatchingEvent(request_event);
    if (duplicate != null && requestsExactlyMatch(request_event, duplicate))
    {
      log.debug("new request was an exact duplicate "
          + "of a current request; dropping");
      return;
    }

    int res = mConcurrentRequestMap.add(request_event, this);

    if (res == ConcurrentRequestMap.DUPLICATE
        || res == ConcurrentRequestMap.DEADLOCK)
    {
      log.trace("new request was a duplicate, "
          + "attaching to in-progress request.");
      return;
    }

    log.trace("processing request");
    processRequest(request_event);
  }

  /**
   * Check to see if two requests exactly match. Exactly matching means that
   * both requests have the same question, same ID, and came from the same
   * source.
   * 
   * @param a One of the request events to check.
   * @param b One of the request events to check.
   * @return true if a and b exactly match, by the defintion above, false if
   *         not.
   */
  private boolean requestsExactlyMatch(DNSEvent a, DNSEvent b)
  {
    Request a_req = a.getOrigRequest();
    Request b_req = b.getOrigRequest();
    if (!a_req.equals(b_req)) return false;

    ServerEventState a_state = getModuleState(a);
    ServerEventState b_state = getModuleState(b);

    if (!a_state.getSourceAddress().equals(b_state.getSourceAddress()))
      return false;

    return true;
  }

  /**
   * Given a response message, validate its channel and match it with an
   * outstanding request. If it came in on a channel not listening for
   * responses, or doesn't correlate with an active request, drop it.
   * Otherwise, add it to the request event and send it along.
   * 
   * @param m The response message.
   * @param channel The channel the response came in on.
   * @param sourceAddress The source address the message came from (needed for
   *          UDP).
   * @param state The state attached to the incoming channel.
   * @throws IOException
   */
  private void handleReadResponse(Message m, SelectableChannel channel,
      InetSocketAddress sourceAddress, ChannelState state) throws IOException
  {
    clientTrafLog.debug("received response from " + sourceAddress + ":\n" + m);
    log.trace("received response from " + sourceAddress + ":\n" + m);

    // First determine if this response came on a proper channel.
    if (!mClientChannelSet.contains(channel))
    {
      // Drop responses arriving on request-only channels.
      log.info("response came on request-only channel; dropping");
      return;
    }

    boolean close_channel = mNSClient.handleResponse(m, sourceAddress);

    // If this response was ready from a non-permanent, "outbound" channel,
    // then it is done (these channels write one request and read one response
    // before closing).
    if (close_channel && !state.isPermanent() && state.isOutbound())
    {
      log.trace("closing outbound channel: " + channel);
      deregisterChannel(channel);
    }
  }

  /**
   * Handle a channel that is writable (and, presumably, has things to write).
   * This will attempt to write as much as it can over the channel.
   * 
   * @param key The selection key in the writable state.
   * @param state The channel state attached to the key.
   * @throws IOException
   */
  private void handleWrite(SelectionKey key, ChannelState state)
      throws IOException
  {
    WriteEvent w = null;
    SelectableChannel channel = key.channel();
    boolean complete;

    if (channel instanceof DatagramChannel)
    {
      DatagramChannel dch = (DatagramChannel) channel;
      while ((w = state.nextWriteEvent()) != null)
      {
        if (w.isClosed())
        {
          // These are normally additional queries generated by wakeups.
          // A lot of these probably means that we are too impatient and
          // waking up too soon.
          log.trace("not writing closed write event");          
          continue;
        }
        complete = mServerUtils.writeUDP(w, dch, state);
        if (!complete)
        {
          log.debug("didn't complete a UDP write.");
          break;
        }

        if (w.isRequest())
        {
          mNSClient.registerWrite(w.getEvent(), w.getMatchState());
        }
      }
    }
    else
    {
      SocketChannel sch = (SocketChannel) channel;
      while ((w = state.nextWriteEvent()) != null)
      {
        if (w.isClosed())
        {
          log.debug("Not sending write event because "
              + "underlying event is closed.");
          continue;
        }

        complete = mServerUtils.writeTCP(w, sch, state);
        if (!complete) break;

        if (w.isRequest())
        {
          mNSClient.registerWrite(w.getEvent(), w.getMatchState());
        }
      }
    }

    // Reset the selection ops back to reading.
    // Note that if this is a channel that only needs to write, it will be
    // closed at the end of this method, so setting the interestOpts to
    // OP_READ is harmless.
    key.interestOps(SelectionKey.OP_READ);

    // If we've gotten here, then we've exhaused the (current) write
    // operations for this key. Check to see if the channel needs closing.
    if (!state.hasWriteEvents() && !state.isPermanent()
        && !state.isOutbound())
    {
      log.trace("closing inbound channel: " + channel);
      deregisterChannel(channel);
    }
  }

  /**
   * The is a convenience method for creating request DNSEvents from DNS
   * messages.
   * 
   * @param m The request message to start with.
   * @param sourceChannel the channel that the request arrived on.
   * @param sourceAddress the address that the request arrived from.
   * 
   * @return The constructed DNSEvent.
   */
  private DNSEvent createRequestEvent(Message m,
      SelectableChannel sourceChannel, InetSocketAddress sourceAddress)
  {
    // Get the source address from the channel, if necessary.
    if (sourceAddress == null && sourceChannel instanceof SocketChannel)
    {
      Socket sock = ((SocketChannel) sourceChannel).socket();
      sourceAddress = (InetSocketAddress) sock.getRemoteSocketAddress();
    }

    ServerEventState state = new ServerEventState(sourceChannel,
        sourceAddress);
    int max_payload = mServerUtils.getMaximumPayloadSize(m,
        sourceChannel instanceof SocketChannel);
    state.setMaximumPayload(max_payload);

    Request request = new Request(m);

    DNSEvent event = new DNSEvent(request);
    event.setModuleState(this, state);

    return event;
  }

  /**
   * Attempt to parse a raw DNS message.
   * 
   * @param in The binary DNS message.
   * @return a Message object if parsing succeeded, null if parsing failed.
   */
  private Message parseMessage(byte[] in)
  {
    try
    {
      Message m = new Message(in);
      return m;
    }
    catch (IOException e)
    {
      log.debug("Problem reading and/or parsing a request message", e);
      log.debug(hexdump.dump("Unparseable DNS message", in, 0, in.length));
    }
    return null;
  }

  /**
   * Write data (a wire format DNS message) to a given destination via a given
   * channel. If the channel is a (connected) TCP channel, the destination is
   * ignored. This method doesn't actually do the I/O -- it really just
   * creates a ChannelState.WriteEvent and schedules the data to be written
   * (by handleWrite()).
   * 
   * @param data The data to send.
   * @param destination The destination to send the data to.
   * @param channel The channel to use (defines source address/port)
   * @param event The event that is trying to write this data. This should
   *          only be supplied when writing a request, not when writing a
   *          response.
   * @param requestMessage The non-wire-format version of data, when this is
   *          used for writing a request.
   */
  private void writeData(byte[] data, InetSocketAddress destination,
      SelectableChannel channel, DNSEvent event, Object matchState)
  {
    SelectionKey key = channel.keyFor(mSelector);
    ChannelState state = (ChannelState) key.attachment();

    state.addWriteEvent(data, destination, event, matchState);
    // FIXME: if TCP channel is still trying to connect, do not set OP_WRITE ?
    key.interestOps(key.interestOps() | SelectionKey.OP_WRITE);
  }

  /**
   * Write data (a wire format DNS message) to a given destination using a
   * (usually) new channel. That is, this message is being sent to a new
   * destination.
   * 
   * @param data The data to send.
   * @param destination The destination to send the data to.
   * @param event The originating event, if this is a request, null otherwise.
   * @param requestMessage The pre-wire-format request message, if this is a
   *          request. null otherwise.
   * @param useTCP if true, then the message will be sent via TCP.
   * @throws IOException
   */
  private void writeData(byte[] data, InetSocketAddress destination,
      DNSEvent event, Object matchState, boolean useTCP) throws IOException
  {
    // log.debug("writeData: outbound query to: " + destination);
    SelectableChannel channel;
    if (mQuerySource != null && !useTCP)
    {
      channel = mQuerySourceChannel;
    }
    else
    {
      if (useTCP)
      {
        SocketChannel sch = SocketChannel.open();
        sch.connect(destination);
        registerTCPChannel(sch, false, true);
        channel = sch;
      }
      else
      {
        DatagramChannel dch = registerNewUDPChannel(null, false, true);
        channel = dch;
      }
    }

    writeData(data, destination, channel, event, matchState);
  }

  /**
   * Write a DNS (response) message. More accurately, this just puts the
   * message in wire format and schedules the message to be written.
   * 
   * @param m The message to send.
   * @param destination The destination (will be ignored if TCP is being used)
   * @param channel The channel to send the message on.
   */
  private void writeResponse(Message m, int max_payload,
      InetSocketAddress destination, SelectableChannel channel)
  {
    byte[] out;

    if (max_payload > 0)
    {
      out = m.toWire(max_payload);
    }
    else
    {
      out = m.toWire();
    }

    log.trace("write message (" + destination + ", "
        + ServerUtils.channelToString(channel) + "): " + out.length
        + " bytes ->\n" + m);
    writeData(out, destination, channel, null, null);
  }

  /**
   * Schedule an outbound request to be written to the target stored in the
   * DNSevent. The request is put in wire format and is added to the outbound
   * request tracking data structure before scheduling.
   *
   * @param out The actual data to write (that is, the wire formatted request).
   * @param destination the IP address and port to send the request to.
   * @param event The request event.
   * @param matchState The match state for this write event.
   * @param useTCP If true, then this request will be sent via TCP.
   */
  public void writeRequest(byte[] out, InetSocketAddress destination,
      DNSEvent event, Object matchState, boolean useTCP)
  {
    try
    {
      writeData(out, destination, event, matchState, useTCP);
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }

  public void logRequest(Message out, InetSocketAddress destination) 
  {
    clientTrafLog.debug("sending request to " + destination +"\n" + out);
  }
  
  /**
   * The main method. This will parse command line options, parse the config
   * file(s), and serve.
   * 
   * @param args Command line arguments, if any.
   */
  public static final void main(String[] args)
  {
    ServerCLIState state = new ServerCLIState();

    try
    {
      state.parseCommandLine(args);

      Server server = new Server();
      server.init(state.config);

      server.serve();

    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

}
